fixed_events:-
  (conjunction(N, _, _, P_Event); implication(N, _, _, P_Event)),
  init_one_fixed_event(N, P_Event, 0),
  fail.
fixed_events.

calc_type(a, 0).
calc_type(cor, 1).
calc_type(v, 2).
calc_type(m, 3).
calc_type(ob, 4).
calc_type(in, 5).
calc_type(event, 6).

init_one_fixed_event(N, event( Equal(Type(Atom),Value) | Tail), Count):-
  Negation := true_value(Equal = '=\='),
  calc_type(Type, Type2),
  encode_cell(Cell, Negation, Value, Type2, Atom),
  event_cell(N,Count) ::= Cell,
  init_one_fixed_event(N, event(|Tail), Count+1).
init_one_fixed_event(N, event(), Count):-
  event_cell(N,Count) ::= max_integer.

fixed_models:-
  fixed_models_data(Model, states(AllStates), Algorithm, Arrows, Traces, Links, StatesTrace),
  all_arrows(Model) := list_length(Arrows), % for arrows, from 1 to 4
  all_traces(Model) := list_length(Traces), % for trace,  from 0 to G_Tr=10
  init_algorithm(Model, Algorithm),
  init_fixed_events(conjunction, Model, 0, 0, Arrows),
  init_fixed_events(implication, Model, 0, 4, Traces),
  init_fixed_Links(Model, Links),
  init_fixed_States(Model, AllStates),
  init_fixed_Trace(Model, StatesTrace),
  fail.
fixed_models.

init_algorithm(Model, Algorithm):-
  (Algorithm=[Enter, Exit]->
    conjunction(Event1, Enter, _, _),
    enter(Model):=Event1,
    conjunction(Event2, Exit, _, _),
    exit(Model):=Event2
  ).

init_fixed_States(Model, AllStates):-
  for(I, 0, AllStates-1),
    for(J, 0, G_Tr-1),
      states(Model,I,J) ::= 5,
      fail. 
init_fixed_States(Model, AllStates).

init_fixed_events(Type, Model, Event, Start, [Ind | List] ):-
  Type(N, Ind, Name, P_Event),
  events(Model,Event+Start) ::= N,
  init_fixed_events(Type, Model, Event+1, Start, List).
init_fixed_events(Type, Model, Event, Start, []).

init_fixed_Trace(Model, [trace(State, Event, Trace) | List] ):-
  states(Model,State,Event) ::= Trace,
  init_fixed_Trace(Model, List).
init_fixed_Trace(Model, []).

init_fixed_Links(Model, [link(State, State2, Event) | List] ):-
  links(Model,State,State2) ::= links(Model,State,State2) \/ 2**Event,
  init_fixed_Links(Model, List).
init_fixed_Links(Model, []).

%  fixed_models_data(Model, states(AllStates), [Enter, Exit], Arrows, Traces, Links, StatesTrace),
% event(p( Atom, Type, Value, Negation)|Tail)
% link(State, State2, Event)
% trace(State, Event, TraceVol)

init_board:-
  for(I, 0, 7),
    init_squre(0, I, I+2),
    init_squre(1, I, 1+14),
    fail.
init_board:-
  for(I, 0, 2),
    init_squre(0, 7-I, I+2),
    fail.
init_board:-
  init_squre(0, 0, 2+14),
  init_squre(0, 4, 6+14),
  init_squre(0, 7, 2+14).

init_squre(Line, Column, Piece):-
    board(Column,Line) ::= Piece,
    board(Column,7-Line) ::= Piece+7.

make_chess_squre(X, I, J):-
  (board(I,J) =:= 7 ->
    color_text_back(_, rgb(255,255,0))
  else ((I+J) mod 2 =:=0, I<8 ->
    color_text_back(_, rgb(0,255,255))
  else
    color_text_back(_, rgb(240,240,240))
  )),
  (board(I,J) mod 14 > 6 ->
    color_text(_, rgb(0,0,0))
  else
    color_text(_, rgb(255,0,0))
  ),
  Piece := board(I,J) mod 7,
  (I =:= 8, active(3) =:= 1 -> Piece := 0),
  face(Piece, Face),
  (I=:=0 -> 
    text_out(Face, pos(G_Zoom_Font*X, G_Zoom*(925-25*J)))
  else
    text_out(Face)
  ).

show_grid(I, J):-
  pen(3, rgb(0, 0, 0)),
  font(G_Zoom_Font*8, G_Zoom_Font*22, "Courier"), 
  X := G_Zoom*(300 - 35*I),
  Y := G_Zoom*(400 - 35*J),
  R := 70*G_Zoom,
  Step := 20*G_Zoom,
  color_text(_, rgb(0,0,255)),
  horizontal_grid(X, Y, I, J, R),
  vertical_grid(X, Y, I, J, R),
  color_text(_, rgb(0,0,0)),
  for(B, 0, J-1),
    for(A, 0, I-1),
      (G_Sel =:= 5 ->
        Object := board(A,7-B)
      else
        Object := board(8,1-B)
      ),
      Object > 0,
      (Object =:= 7 ->
        color_text(_, rgb(0,255,0)),
        names(24, _, _, _, Name),
        text_out(Name, pos(2+X+A*R, 3+Y+B*R)),
        fail
      else (Object mod 14 > 6 ->
        names(17, _, _, _, Name),
        color_text(_, rgb(0,0,0))
      else
        names(16, _, _, _, Name),
        color_text(_, rgb(255,0,0))
      )),
      text_out(Name, pos(2+X+A*R, 3+Y+B*R)),
      color_text(_, rgb(0,0,255)),
      Piece := 17+ Object mod 7,
      names(Piece, _, _, _, Name2),
      text_out(Name2, pos(2+X+A*R, 3+Step+Y+B*R)),
      Object > 14,
      color_text(_, rgb(0,0,0)),
      names(25, _, _, _, Name3),
      text_out(Name3, pos(2+X+A*R, 3+2*Step+Y+B*R)),
      fail.
show_grid(I, J):-
  font(G_Zoom_Font*10, G_Zoom_Font*25, "Courier"). 

horizontal_grid(X, Y, I, J, R):-
  for(B, 0, J),
    line(X, Y+B*R, X+I*R, Y+B*R),
    B < J, J > 1,
    text_out(J-B, pos(X-G_Zoom*17, G_Zoom*25+Y+B*R)),
    fail.
horizontal_grid(X, Y, I, J, R).

vertical_grid(X, Y, I, J, R):-
  for(A, 0, I),
    line(X+A*R, Y, X+A*R, Y+J*R),
    A < I, I > 1,
    text_out(A+1, pos(X+G_Zoom*30+A*R, Y+G_Zoom*5+J*R)),
    fail.
vertical_grid(X, Y, I, J, R).

observable(Atom) :-
  (Atom > 6, Atom < 26 ->
    (G_Observed > 0 ->
      Atom =:= G_Observed
    else (Atom > 15 ->
      here(I, J),
      Squre := board(I,J),
      Squre > 0,
      (Atom < 18 -> % colour
        (Squre mod 14 > 6 -> Atom=:=17, Squre=\=7 else Atom=:=16)
      else (Atom < 24 -> % piece
        Atom - 17 =:= Squre mod 7
      else
        (Atom=:=24, Squre =:=7; Atom=:=25, Squre > 14)
      ))
    else
      active(Atom)=\=0
    ))
  ).

up:- % m0=2 & a0=a
  active(0) =:= 2**0, % 0 but previous was 2
  buf_actions(G_Time mod 50)=:=1.

down :- % m0=2 & a0=b
  active(0) =:= 2**0,
  buf_actions(G_Time mod 50)=:=2.

here(I, J):- % the same as previous (after up and down)
  I := find_first_1_in(active(1)),
  J := find_first_1_in(active(2)).

somewhere(I, J):-
  repeat,
  I := random(8),
  J := random(8).

update_memory :- % up, here => copy(Lifted)
  up,
  here(I, J),
  board(I,J) ::= 7,
  fail.
update_memory :- % down, everywhere => del(Lifted)
  down,
  for(I, 0, 7),
    for(J, 0, 7),
      (board(I,J) =:= 7 -> board(I,J) ::= 0),
      fail.
update_memory :- % down, here => copy(mem7(0))
  down,
  here(I, J),
  board(I,J) ::= board(8,0),
  fail.
update_memory :- % down, !ob(Lifted), here => del(Unmoved)
  down,
  G_ObserveNow =\= 7, % the previous value
  here(I, J),
  board(I,J) ::= board(I,J) mod 14,
  fail.
update_memory :- %  => ob(now) := mem6(here)
  here(I, J),
  G_ObserveNow := board(I,J),
  fail.
/*----------------*/
update_memory :- %  here = 1 => mem7(here) := ob(now)
  0 =:= find_first_1_in(active(3)),
  board(8,0) ::= G_ObserveNow,
  fail.
update_memory.

show_memory_rules:-
  I := 2,
  memory_rules(Number, G_Sel, Text1, Text2),
  I := I + 1,
  color_text(_, rgb(255,0,0)),
  text_out(Number, pos(G_Zoom*600, G_Zoom*(30+I*40))),
  color_text(_, rgb(0,0,0)),
  text_out(Text1),
  color_text(_, rgb(0,0,255)),
  text_out(Text2).

evaluete_events(_, _):- % transition and ...
  conjunction(Event, Ind, Name, P_Event),
  Cell := event_cell(Event,0),
  (Cell =:= max_integer ->
    events_val(Event) := 5
  else (event_is_true(Event, 0, 0) -> 
    decode_cell(Cell, _, _, Type, _),
    events_val(Event) := (Type =:= 2 -> 3 else 5)
  else
    events_val(Event) :=0
  )),
  fail.
evaluete_events(_, forcast):-
  implication(Event, Ind, Name, P_Event),
  Cell := event_cell(Event,0),
  decode_cell(Cell, Negative, Value, Type, _),
  (Cell =:= max_integer ->
    write("Error"), nl
  else (event_is_true(Event, 1, 0), Type<3, is_forcast(Event, Type)->
    (Type>0 ->
      events_val(Event) := encode_events_val(Negative, Value, 3-Type, 0)
    else
      events_val(Event) := encode_events_val(Negative, Value, 3-   1, 1)
    )
  else
    events_val(Event) :=0
  )),
  fail.
evaluete_events(Views, restriction):-
  implication(Event, Ind, Name, P_Event),
  events_val(Event) :=0,
  Cell := event_cell(Event,0),
  decode_cell(Cell, _, _, Type, _),
  (Views=:=1 ->
    Type =:= 2
  else (Views=:=3 ->
    Type < 2
  else
    Type > 2
  )),
  (Cell =:= max_integer ->
    write("Error"), nl
  else (event_is_true(Event, 1, 0)->
    events_val(Event) := 2 + true_value(cell_is_true(Cell, 0))
  )),
  fail.
evaluete_events(_, _).

is_forcast(Event, Type):-
  Type1 := event_cell(Event,1) //G_S//2, % Type + 10*Atom
  (Type=:=2 -> Type1=\=2 else Type1>1).

events_are_true(Model, Events, Views):-
  Events := 0,
  for(Event, 0, all_arrows(Model)-1),
    Event2 := events(Model,Event),
    (events_val(Event2) =:= 4-Views ->
      Events := Events \/ 2**Event
    ),
    fail.
events_are_true(Model, Events, Views).

event_is_true(Event, Count, Delay):-
  Cell := event_cell(Event,Count),
  (Cell =\= max_integer ->
    cell_is_true(Cell, Delay),
    event_is_true(Event, Count+1, Delay)
  ).

cell_is_true(Cell, Delay):-
  decode_cell(Cell, Negative, Value, Type, Atom),
  (Type =:= 1 ->
    Res := true_value(buf_incorrect((G_Time+Delay-Atom) mod 50) /\ 2**Value =:= 0)
  else (Type =:= 3 ->
    Res := true_value(active(Atom) /\ 2**Value =\= 0)
  else (Type =:= 4 ->
    Res := true_value(observable(Atom))
  else (Type =:= 6 ->
    Res := events_val(Atom)
  else
    Buffer := (Type =:= 0 -> buf_actions else buf_views),
    Res := true_value(Buffer((G_Time+Delay-Atom) mod 50) =:= Value)
  )))),
  (Res + Negative) mod 2 =:= 1.

R is links(int X, int Y, int Z) :- R := links3(G_S*(G_S*X+Y)+Z).
R is states(int X, int Y, int Z) :- R := states3(G_Tr*(G_S*X+Y)+Z).
R is events(int X, int Y) :- R := events2(G_T*X+Y).
R is event_cell(int X, int Y) :- R := event_cell2(10*X+Y).
R is results(int X, int Y, int Z) :- R := results3(32*X+4*Y+Z).
R is res(int X, int Y) :- R := res2(4*X+Y).
R is active_last(int X, int Y) :- R := active_last2(2*X+Y).
R is events_last(int X, int Y) :- R := events_last2(2*X+Y).
R is buf_arrows(int X, int Y) :- R := buf_arrows2(50*X+Y).
R is buf_the_trace(int X, int Y) :- R := buf_the_trace2(50*X+Y).
R is board(int X, int Y) :- R := board2(8*X+Y).
R is statistic(int X, int Y, int Z) :- R := statistic3(72*X+8*Y+Z).
R is uniqueness(int X, int Y, int Z) :- R := uniqueness3(72*X+8*Y+Z).

links(X, Y, Z) ::= Val :- links3(G_S*(G_S*X+Y)+Z) := Val.
states(X, Y, Z) ::= Val :- states3(G_Tr*(G_S*X+Y)+Z) := Val.
events(X, Y) ::= Val :- events2(G_T*X+Y) := Val.
event_cell(X, Y) ::= Val :- event_cell2(10*X+Y) := Val.
results(X, Y, Z) ::= Val :- results3(32*X+4*Y+Z) := Val.
res(X, Y) ::= Val :- res2(4*X+Y) := Val.
active_last(X, Y) ::= Val :- active_last2(2*X+Y) := Val.
events_last(X, Y) ::= Val :- events_last2(2*X+Y) := Val.
buf_arrows(X, Y) ::= Val :- buf_arrows2(50*X+Y) := Val.
buf_the_trace(X, Y) ::= Val :- buf_the_trace2(50*X+Y) := Val.
board(X, Y) ::= Val :- board2(8*X+Y) := Val.
statistic(X, Y, Z) ::= Val :- statistic3(72*X+8*Y+Z) := Val.
uniqueness(X, Y, Z) ::= Val :- uniqueness3(72*X+8*Y+Z) := Val.
