% This program is created by students
% from University of Sofia

?-
  init,
  G_Win := window(title("Digital TV"), size(500, 300), pos(100, 100), not(clean_before_paint) ),
           window(title("Parameters"), size(500, 300), pos(600, 100), class(win_func2), parent(G_Win), paint_indirectly).

[X | Tail] :== [A | T1] + S*[B | T2] :-
  X := A + S*B,
  Tail :== T1 + S*T2.
[] :== [] + S*[].

calc_pixel(I, J, Color) :- 
  calc_pull(Pull, I, J, A, B),
  C := sqrt(A**2 + B**2),
  C =\= 0, 
  Sin := A/C,
  Cos := B/C,
  calc_cubic_root(Sin, Cos, 1, I, J, Dist),
  G_Sign * Dist > 0,
  T := (G_Sign * Dist / Pull)/175,
  Color1 := 137 - ty(T, 14),
  D := 1.05**G_Contrast * Color1 + 5*G_Brightness,
  (D > 255 ->
    Color := rgb(255, 128, 0)
  else (D < 0 ->
    Color := rgb(0, 128, 255)
  else
    Color := rgb(D, 128, 255-D) % rgb(D, D, D) for Black & White
  )).

calc_pull(Pull, I, J, A, B) :- 
  (G_Y =:= 200, y(0)=\=200 ->
    A := x(0) - x(1), 
    B := y(0) - y(1)
  else
    A := I - G_X, 
    B := G_Y - J
  ),
  (B =:= 0 ->
    X := 225
  else 
    X := I + (J - 200)*A/B
  ),
  X > 150, X < 300,
  T := (X-150)/150,
  Pull := (200 - ty(T, 10))/100.

calc_center :-
  calc_abc(A1, B1, C1, 0, 1), 
  calc_abc(A2, B2, C2, 5, 4),
  G_Sign := (x(5)-x(0)> x(4)-x(1) -> -1 else 1),
  (A1*B2 =:= A2*B1 ->
    G_X := (x(0)+x(5))/2,
    G_Y := 200
  else (B1 =:= 0 ->
    G_X := -1*C1/A1,
    G_Y := -1*(A2*G_X+C2)/B2
  else 
    G_X := (C1*B2/B1-C2)/(A2-A1*B2/B1),
    G_Y := -1*(A1*G_X+C1)/B1
  )).

calc_abc(A, B, C, P1, P2):-
  A := y(P1)-y(P2),
  B := x(P2)-x(P1),
  C := -1*x(P2)*A-y(P2)*B.

win_func(init) :- 
  check_box(checked, pos(10, 220), size(80, 20), text("Borders"), class(check_func)).

win_func(paint) :- 
  for(I, 0, 470), 
    for(J, 0, 250),
      (calc_pixel(I, J, Color) -> true else Color := system_color(window)),
      draw_point(I, J, Color),
      fail.

win_func(paint) :-
  G_Borders =:= 1,
  make_borders(1).

make_borders(N) :-
  pen(3, rgb(255, 0, 0)),
  bezier_line( x(1), y(1), x(2), y(2), x(3), y(3), x(4), y(4)),
    pen(1, rgb(0, 0, 255)),
    line(x(0), y(0), x(1), y(1)),
    line(x(4), y(4), x(5), y(5))
  ,
  make_arc(N). 
make_borders(N).
  
make_arc(N):-
  for(T0, 0, 1000),
    T is T0/1000,
    X := tx(T, 1),
    Y := ty(T, 1),
    calc_pull(Pull, X, Y, A, B),
    [Xp, Yp] :== [X, Y]+ -1*[G_X, G_Y],
    C := -50*Pull*G_Sign/sqrt(Xp**2 + Yp**2),
    [X2, Y2] :== [X, Y]+  C*[Xp, Yp],
    [X3, Y3] :== [X, Y]+  (2*C)*[Xp, Yp],
    [X4, Y4] :== [X, Y]+  (3*C)*[Xp, Yp],
      draw_point(X2, Y2, rgb(0,0,255)),
      N=:=1,
      draw_point(X3, Y3, rgb(0,0,255)),
      draw_point(X4, Y4, rgb(0,0,255)),
  fail.

check_func(press) :-
  G_Borders := 1-G_Borders,
  set_check_box_value(_, G_Borders),
  (G_Borders =:= 1 ->
    make_borders(1)
  else
    update_window(_)
  ).

update_func(press) :-
  update_window(G_Win).

set_rb(N):-
  G_Page := N, 
  set_check_box_value(G_RB1, 0),
  set_check_box_value(G_RB2, 0), 
  set_check_box_value(G_RB3, 0),
  set_check_box_value(G_RB4, 0), 
  set_check_box_value(_, 1),
  update_window(_).

rb1(press) :- set_rb(1).
rb2(press) :- set_rb(2).
rb3(press) :- set_rb(3).
rb4(press) :- set_rb(4).

win_func2(init) :- 
  button( pos(380, 220), size(80, 20), text("Update"), class(update_func)),
  G_RB1 := radio_button(checked, pos(10,  10), size(80, 20), text("Main"), class(rb1)),
  G_RB2 := radio_button(         pos(130, 10), size(80, 20), text("Pull"), class(rb2)),
  G_RB3 := radio_button(         pos(250, 10), size(80, 20), text("Color"), class(rb3)),
  G_RB4 := radio_button(         pos(370, 10), size(80, 20), text("Brightness"), class(rb4)).

win_func2(paint) :-
  paint_page(G_Page),
  pen(3, rgb(0, 0, 0)),
  brush(rgb(255,255,255)), 
  movable(P, G_Page),
  ellipse(x(P)-10, y(P)-10, x(P)+10, y(P)+10),
  fail.

win_func2(mouse_click(X, Y)) :-
  movable(P, G_Page),
  close_to(X, Y, P),
  G_P := P,
  update(X, Y).

win_func2(mouse_move(X, Y)) :-
  G_P > -1, 
  update(X, Y).

win_func2(mouse_click_up(X, Y)) :-
  G_P := -1.

win_func2(mouse_wheel(Delta, X, Y)):-
  G_Page > 2,
  (message_flags(left); message_flags(ctrl)->
    G_Contrast := max(-50, min(50, G_Contrast + Delta))
  else
    G_Brightness := max(0, min(100, G_Brightness + Delta))
  ),
  update_window(_).

win_func2(close) :- 
  close(G_Win).

movable(P, 1):- member(P, [0, 2, 3, 5]).
movable(P, 2):- member(P, [10, 11, 12, 13]).
movable(P, 3):- member(P, [14, 15, 16, 17]).

paint_page(1):- % Main
  pen(3, rgb(0, 128, 0)),
  line(x(1), y(1), x(2), y(2)),
  line(x(3), y(3), x(4), y(4)),
  pen(1, rgb(0, 0, 255)), 
  line(x(0), y(0), x(5), y(5)),
  pen(0, 0), 
  brush(rgb(0,0,0)), 
  ellipse(G_X-3, G_Y-3, G_X+3, G_Y+3),
  make_borders(1). 

paint_page(2):- % Pull
  make_borders(2),
  pen(1, rgb(0, 128, 0)),
  line(50, 100, 400, 100),
  pen(1, rgb(255, 0, 0)),
  line(50, 200, 400, 200),
  for(T0, 0, 350),
    T is T0/350,
    Y := ty(T, 10),
      draw_point(T0+x(10), Y, rgb(0,128,0)),
  fail.
paint_page(2).

paint_page(3):- % Color
  color_text(_, rgb(0,0,0)),
  text_out("B: "+G_Brightness, pos(10, 50)), 
  text_out("C: "+(50+G_Contrast), pos(10, 80)),
 
  Br := G_Brightness+119,
  Contr := 128/1.05**G_Contrast,
  pen(0, 0),
  brush(rgb(255,255,0)),
  rect(50, Br-Contr, 400, Br+Contr),
  pen(1, rgb(0, 0, 0)),
  line(50, Br-Contr, 400, Br-Contr),
  line(50, Br+Contr, 400, Br+Contr),
  line(50, 50, 50, 200),
  pen(3, rgb(0, 0, 255)),
  line(150, 50, 150, 200),
  line(250, 50, 250, 200),
  line(350, 50, 350, 200),
  for(T0, 0, 1000),
    T is T0/350,
    Y := ty(T, 14),
      draw_point(T0+x(14), Y, (T>1-> rgb(0,128,0) else rgb(255,0,255))),
  fail.
paint_page(3).

paint_page(4):- % Brightness
  color_text(_, rgb(0,0,0)),
  text_out("Brightness "+G_Brightness, pos(10, 50)), 
  color_text(_, rgb(255,0,0)),
  text_out(" (mouse wheel)"), 
  color_text(_, rgb(0,0,0)),
  text_out("Contrast "+(50+G_Contrast), pos(10, 80)), 
  color_text(_, rgb(255,0,0)),
  text_out(" (ctrl + mouse wheel)"), 
  pen(3, rgb(0, 0, 0)),
  brush(rgb(255,255,0)),
  rect(300, G_Brightness+100-10/1.05**G_Contrast, 450, G_Brightness+100+10/1.05**G_Contrast).

update(X, Y) :- 
  (G_P < 10 ->
    x(G_P) := X
  ), 
  y(G_P) := Y,
  (G_P =:= 0 -> y(5) := Y),
  (G_P =:= 5 -> y(0) := Y),
  (G_P =:= 0; G_P =:= 5->
    calc_center
  ),
  update_window(_).

close_to(X, Y, P) :-
  (X - x(P))**2 + (Y - y(P))**2 < 100.

init :- 
  G_Brightness := 30,
  G_Contrast := 20, % +50
  G_P := -1,
  G_Borders := 1,
  G_Page := 1, 
  array(x, 20, 0.0),
  array(y, 20, 0.0),
  init_xy(0, [p(50, 150), p(150, 200), p(150, 100), p(300, 100), p(300, 200), p(400, 150)]),
  init_xy(10, [p(50, 100), p(150, 100), p(300, 100), p(400, 100)]),
  init_xy(14, [p(50, 200), p(150, 170), p(300, 130), p(400, 100)]),
  calc_center. 

init_xy(From, [p(X, Y) | Tail]) :- 
  x(From) := X,
  y(From) := Y,
  init_xy(From +1, Tail).

init_xy(_, []).

calc_cubic_root(Sin, Cos, Point, I, J, Distance):-
    X0 is Cos*x(Point) + Sin*y(Point),
    X1 is Cos*x(1+Point) + Sin*y(1+Point),
    X2 is Cos*x(2+Point) + Sin*y(2+Point),
    X3 is Cos*x(3+Point) + Sin*y(3+Point),
    Ir is Cos*I + Sin*J,
    Jr is Cos*J - Sin*I,
    A is -1*X0+3*X1-3*X2+X3,
    A =\= 0,
    B is  3*X0-6*X1+3*X2,
    C is -3*X0+3*X1,
    D is    X0-Ir,
    %Discrim is 18*A*B*C*D - 4*B**3*D + B**2*C**2 - 4*A*C**3 - 27*A**2*D**2,
    Dis0 is B**2 - 3*A*C,
    Dis1 is 2*B**3 - 9*A*B*C + 27*A**2*D,
    Dis is (Dis1**2 - 4*Dis0**3),
    (Dis > 0 ->
      Temp1 is (Dis1+sqrt(Dis))/2,
      Temp2 is (Dis1-sqrt(Dis))/2,
      Temp is (abs(Temp1)>abs(Temp2)-> Temp1 else Temp2),
      Const is (Temp>0-> Temp**(1/3) else -1*(-1*Temp)**(1/3)),
      %Const is ((Dis1+sqrt(Dis2))/2)**(0.333),
      (Const=:=0.0-> write([Const, Dis0, Dis1, Dis, Temp]), nl),
      T  is (B+Const+Dis0/Const)/(-3*A),
      Distance := Jr-fy(T, Sin, Cos, Point)
    else
      R0 is sqrt((Dis1**2-Dis)/4),
      R  is R0**(1/3),
      Fi is arccos(Dis1/(2*R0))/3,
      Ta is (B+cos(Fi)       *(R+Dis0/R))/(-3*A),
      Tb is (B+cos(Fi+2*pi/3)*(R+Dis0/R))/(-3*A),
      Tc is (B+cos(Fi+4*pi/3)*(R+Dis0/R))/(-3*A),
      Ya is Jr-fy(Ta, Sin, Cos, Point),
      Yb is Jr-fy(Tb, Sin, Cos, Point),
      Yc is Jr-fy(Tc, Sin, Cos, Point),
      Distance := -1000,
      (Ta>=0, Ta=<1              -> Distance := Ya),
      (Tb>=0, Tb=<1, Distance<Yb -> Distance := Yb),
      (Tc>=0, Tc=<1, Distance<Yc -> Distance := Yc)
    ).

Y is fy(T, Sin, Cos, P):-
  T1 is 1-T,
  Y := T1**3*(Cos*y(P)-Sin*x(P)) + 3*T1**2*T*(Cos*y(1+P)-Sin*x(1+P)) + 3*T1*T**2*(Cos*y(2+P)-Sin*x(2+P)) + T**3*(Cos*y(3+P)-Sin*x(3+P)).

X is tx(T, P):-
  T1 is 1-T,
  X := T1**3*x(P) + 3*T1**2*T*x(P+1) + 3*T1*T**2*x(P+2) + T**3*x(P+3).

Y is ty(T, P):-
  T1 is 1-T,
  Y := T1**3*y(P) + 3*T1**2*T*y(P+1) + 3*T1*T**2*y(P+2) + T**3*y(P+3).
